@@ -1,3 +1,4 @@
+import sys
 import functools
 import itertools
 import logging
@@ -7,12 +8,12 @@ import cvxpy
 import numpy as np
 from sklearn.metrics import mean_absolute_error, median_absolute_error, mean_squared_error
 
-logger = logging.getLogger(__name__)
+logging.basicConfig(stream=sys.stdout, level=logging.INFO, format='%(asctime)s - %(levelname)s - %(message)s')
 
 
 class ASGL:
     def __init__(self, model, penalization, intercept=True, tol=1e-5, lambda1=1, alpha=0.5, tau=0.5,
-                 lasso_weights=None, gl_weights=None, parallel=False, num_cores=None, solver=None, max_iters=500):
+                 lasso_weights=None, gl_weights=None, parallel=False, num_cores=None, solver='default', max_iters=500):
         """
         Parameters:
             model: model to be fit (accepts 'lm' or 'qr')
@@ -27,7 +28,7 @@ class ASGL:
             gl_weights: group lasso weights in adaptive penalizations
             parallel: boolean, whether to execute the code in parallel or sequentially
             num_cores: if parallel is set to true, the number of cores to use in the execution. Default is (max - 1)
-            solver: solver to be used by CVXPY. Default uses optimal alternative depending on the problem
+            solver: solver to be used by CVXPY. default uses optimal alternative depending on the problem
             max_iters: CVXPY parameter. Default is 500
 
         Returns:
@@ -50,15 +51,13 @@ class ASGL:
         self.num_cores = num_cores
         self.max_iters = max_iters
         self.coef_ = None
-        # Define solver param as a list of the three default solvers for CVXPY
-        if solver is None:
-            self.solver = ['ECOS', 'OSQP', 'SCS']
-        else:
-            self.solver = [solver]
+        # CVXPY solver parameters
+        self.solver_stats = None
+        self.solver = solver
 
     # Model checker related functions #################################################################################
 
-    def __model_checker(self):
+    def _model_checker(self):
         """
         Checks if the input model is one of the valid options:
          - lm for linear models
@@ -67,10 +66,10 @@ class ASGL:
         if self.model in self.valid_models:
             return True
         else:
-            logger.error(f'{self.model} is not a valid model. Valid models are {self.valid_models}')
+            logging.error(f'{self.model} is not a valid model. Valid models are {self.valid_models}')
             return False
 
-    def __penalization_checker(self):
+    def _penalization_checker(self):
         """
         Checks if the penalization is one of the valid options:
          - lasso for lasso penalization
@@ -83,11 +82,11 @@ class ASGL:
         if (self.penalization in self.valid_penalizations) or (self.penalization is None):
             return True
         else:
-            logger.error(f'{self.penalization} is not a valid penalization. '
-                         f'Valid penalizations are {self.valid_penalizations} or None')
+            logging.error(f'{self.penalization} is not a valid penalization. '
+                          f'Valid penalizations are {self.valid_penalizations} or None')
             return False
 
-    def __dtype_checker(self):
+    def _dtype_checker(self):
         """
         Checks if some of the inputs are in the correct format
         """
@@ -100,16 +99,16 @@ class ASGL:
         response = response_1 and response_2
         return response
 
-    def __input_checker(self):
+    def _input_checker(self):
         """
         Checks that every input parameter for the model solvers has the expected format
         """
-        response_list = [self.__model_checker(), self.__penalization_checker(), self.__dtype_checker()]
+        response_list = [self._model_checker(), self._penalization_checker(), self._dtype_checker()]
         return False not in response_list
 
     # Preprocessing related functions #################################################################################
 
-    def __preprocessing_lambda(self):
+    def _preprocessing_lambda(self):
         """
         Processes the input lambda1 parameter and transforms it as required by the solver package functions
         """
@@ -123,7 +122,7 @@ class ASGL:
             n_lambda = len(lambda_vector)
         return n_lambda, lambda_vector
 
-    def __preprocessing_alpha(self):
+    def _preprocessing_alpha(self):
         """
         Processes the input alpha parameter from sgl and asgl penalizations and transforms it as required by the solver
         package functions
@@ -139,7 +138,7 @@ class ASGL:
                 n_alpha = len(alpha_vector)
         return n_alpha, alpha_vector
 
-    def __preprocessing_weights(self, weights):
+    def _preprocessing_weights(self, weights):
         """
         Converts l_weights into a list of lists. Each list inside l_weights defines a set of weights for a model
         """
@@ -165,7 +164,7 @@ class ASGL:
                 n_weights = len(weights_list)
         return n_weights, weights_list
 
-    def __preprocessing_itertools_param(self, lambda_vector, alpha_vector, lasso_weights_list, gl_weights_list):
+    def _preprocessing_itertools_param(self, lambda_vector, alpha_vector, lasso_weights_list, gl_weights_list):
         """
         Receives as input the results from preprocessing_lambda, preprocessing_alpha and preprocessing_weights
         Outputs an iterable list of parameter values "param"
@@ -182,35 +181,35 @@ class ASGL:
             param = itertools.product(lambda_vector, alpha_vector, lasso_weights_list, gl_weights_list)
         else:
             param = None
-            logger.error(f'Error preprocessing input parameters')
+            logging.error(f'Error preprocessing input parameters')
         param = list(param)
         return param
 
-    def __preprocessing(self):
+    def _preprocessing(self):
         """
         Receives all the parameters of the models and creates tuples of the parameters to be executed in the penalized
         model solvers
         """
         # Run the input_checker to verify that the inputs have the correct format
-        if self.__input_checker() is False:
-            logger.error('incorrect input parameters')
+        if self._input_checker() is False:
+            logging.error('incorrect input parameters')
             raise ValueError('incorrect input parameters')
         # Defines param as None for the unpenalized model
         if self.penalization is None:
             param = None
         else:
             # Reformat parameter vectors
-            n_lambda, lambda_vector = self.__preprocessing_lambda()
-            n_alpha, alpha_vector = self.__preprocessing_alpha()
-            n_lasso_weights, lasso_weights_list = self.__preprocessing_weights(self.lasso_weights)
-            n_gl_weights, gl_weights_list = self.__preprocessing_weights(self.gl_weights)
-            param = self.__preprocessing_itertools_param(lambda_vector, alpha_vector, lasso_weights_list,
-                                                         gl_weights_list)
+            n_lambda, lambda_vector = self._preprocessing_lambda()
+            n_alpha, alpha_vector = self._preprocessing_alpha()
+            n_lasso_weights, lasso_weights_list = self._preprocessing_weights(self.lasso_weights)
+            n_gl_weights, gl_weights_list = self._preprocessing_weights(self.gl_weights)
+            param = self._preprocessing_itertools_param(lambda_vector, alpha_vector, lasso_weights_list,
+                                                        gl_weights_list)
         return param
 
     # CVXPY SOLVER RELATED OPTIONS ###################################################################################
 
-    def __cvxpy_solver_options(self, solver):
+    def _cvxpy_solver_options(self, solver):
         if solver == 'ECOS':
             solver_dict = dict(solver=solver,
                                max_iters=self.max_iters)
@@ -223,13 +222,13 @@ class ASGL:
 
     # SOLVERS #########################################################################################################
 
-    def __quantile_function(self, x):
+    def _quantile_function(self, x):
         """
         Quantile function required for quantile regression models.
         """
         return 0.5 * cvxpy.abs(x) + (self.tau - 0.5) * x
 
-    def __num_beta_var_from_group_index(self, group_index):
+    def _num_beta_var_from_group_index(self, group_index):
         """
         Internal function used in group based penalizations (gl, sgl, asgl, asgl_lasso, asgl_gl)
         """
@@ -253,24 +252,33 @@ class ASGL:
         if self.model == 'lm':
             objective_function = (1.0 / n) * cvxpy.sum_squares(y - x @ beta_var)
         else:
-            objective_function = (1.0 / n) * cvxpy.sum(self.__quantile_function(x=(y - x @ beta_var)))
+            objective_function = (1.0 / n) * cvxpy.sum(self._quantile_function(x=(y - x @ beta_var)))
         objective = cvxpy.Minimize(objective_function)
         problem = cvxpy.Problem(objective)
-        # Solve the problem. Try first default CVXPY option, which is usually optimal for the problem. If a ValueError
-        # arises, try the solvers provided as input to the method.
+        # Solve the problem. If solver is left as default, try optimal solver sugested by cvxpy.
+        # If other name is provided, try the name provided
+        # If these options fail, try default ECOS, OSQP, SCS options
         try:
-            problem.solve(warm_start=True)
+            if self.solver == 'default':
+                problem.solve(warm_start=True)
+            else:
+                solver_dict = self._cvxpy_solver_options(solver=self.solver)
+                problem.solve(**solver_dict)
         except (ValueError, cvxpy.error.SolverError):
-            for elt in self.solver:
-                solver_dict = self.__cvxpy_solver_options(solver=elt)
+            logging.warning('Default solver failed. Using alternative options. Check solver and solver_stats for more '
+                            'details')
+            solver = ['ECOS', 'OSQP', 'SCS']
+            for elt in solver:
+                solver_dict = self._cvxpy_solver_options(solver=elt)
                 try:
                     problem.solve(**solver_dict)
                     if 'optimal' in problem.status:
                         break
                 except (ValueError, cvxpy.error.SolverError):
                     continue
+        self.solver_stats = problem.solver_stats
         if problem.status in ["infeasible", "unbounded"]:
-            logger.warning('Optimization problem status failure')
+            logging.warning('Optimization problem status failure')
         beta_sol = beta_var.value
         beta_sol[np.abs(beta_sol) < self.tol] = 0
         return [beta_sol]
@@ -295,32 +303,42 @@ class ASGL:
         if self.model == 'lm':
             objective_function = (1.0 / n) * cvxpy.sum_squares(y - x @ beta_var)
         else:
-            objective_function = (1.0 / n) * cvxpy.sum(self.__quantile_function(x=(y - x @ beta_var)))
+            objective_function = (1.0 / n) * cvxpy.sum(self._quantile_function(x=(y - x @ beta_var)))
         objective = cvxpy.Minimize(objective_function + lasso_penalization)
         problem = cvxpy.Problem(objective)
         beta_sol_list = []
         # Solve the problem iteratively for each parameter value
         for lam in param:
             lambda_param.value = lam
-            # Solve the problem. Try first default CVXPY option, which is usually optimal for the problem. If a
-            # ValueError arises, try the solvers provided as input to the method.
+            # Solve the problem. If solver is left as default, try optimal solver sugested by cvxpy.
+            # If other name is provided, try the name provided
+            # If these options fail, try default ECOS, OSQP, SCS options
             try:
-                problem.solve(warm_start=True)
+                if self.solver == 'default':
+                    problem.solve(warm_start=True)
+                else:
+                    solver_dict = self._cvxpy_solver_options(solver=self.solver)
+                    problem.solve(**solver_dict)
             except (ValueError, cvxpy.error.SolverError):
-                for elt in self.solver:
-                    solver_dict = self.__cvxpy_solver_options(solver=elt)
+                logging.warning(
+                    'Default solver failed. Using alternative options. Check solver and solver_stats for more '
+                    'details')
+                solver = ['ECOS', 'OSQP', 'SCS']
+                for elt in solver:
+                    solver_dict = self._cvxpy_solver_options(solver=elt)
                     try:
                         problem.solve(**solver_dict)
                         if 'optimal' in problem.status:
                             break
                     except (ValueError, cvxpy.error.SolverError):
                         continue
+            self.solver_stats = problem.solver_stats
             if problem.status in ["infeasible", "unbounded"]:
-                logger.warning('Optimization problem status failure')
+                logging.warning('Optimization problem status failure')
             beta_sol = beta_var.value
             beta_sol[np.abs(beta_sol) < self.tol] = 0
             beta_sol_list.append(beta_sol)
-        logger.debug('Function finished without errors')
+        logging.debug('Function finished without errors')
         return beta_sol_list
 
     def gl(self, x, y, group_index, param):
@@ -329,8 +347,9 @@ class ASGL:
         """
         n = x.shape[0]
         # Check th group_index, find the unique groups, count how many vars are in each group (this is the group size)
+        group_index = np.asarray(group_index).astype(int)
         unique_group_index = np.unique(group_index)
-        group_sizes, beta_var = self.__num_beta_var_from_group_index(group_index)
+        group_sizes, beta_var = self._num_beta_var_from_group_index(group_index)
         num_groups = len(group_sizes)
         model_prediction = 0
         group_lasso_penalization = 0
@@ -354,7 +373,7 @@ class ASGL:
         if self.model == 'lm':
             objective_function = (1.0 / n) * cvxpy.sum_squares(y - model_prediction)
         else:
-            objective_function = (1.0 / n) * cvxpy.sum(self.__quantile_function(x=(y - model_prediction)))
+            objective_function = (1.0 / n) * cvxpy.sum(self._quantile_function(x=(y - model_prediction)))
         lambda_param = cvxpy.Parameter(nonneg=True)
         objective = cvxpy.Minimize(objective_function + (lambda_param * group_lasso_penalization))
         problem = cvxpy.Problem(objective)
@@ -362,21 +381,31 @@ class ASGL:
         # Solve the problem iteratively for each parameter value
         for lam in param:
             lambda_param.value = lam
-            # Solve the problem. Try first default CVXPY option, which is usually optimal for the problem. If a
-            # ValueError arises, try the solvers provided as input to the method.
+            # Solve the problem. If solver is left as default, try optimal solver sugested by cvxpy.
+            # If other name is provided, try the name provided
+            # If these options fail, try default ECOS, OSQP, SCS options
             try:
-                problem.solve(warm_start=True)
+                if self.solver == 'default':
+                    problem.solve(warm_start=True)
+                else:
+                    solver_dict = self._cvxpy_solver_options(solver=self.solver)
+                    problem.solve(**solver_dict)
             except (ValueError, cvxpy.error.SolverError):
-                for elt in self.solver:
-                    solver_dict = self.__cvxpy_solver_options(solver=elt)
+                logging.warning(
+                    'Default solver failed. Using alternative options. Check solver and solver_stats for more '
+                    'details')
+                solver = ['ECOS', 'OSQP', 'SCS']
+                for elt in solver:
+                    solver_dict = self._cvxpy_solver_options(solver=elt)
                     try:
                         problem.solve(**solver_dict)
                         if 'optimal' in problem.status:
                             break
                     except (ValueError, cvxpy.error.SolverError):
                         continue
+            self.solver_stats = problem.solver_stats
             if problem.status in ["infeasible", "unbounded"]:
-                logger.warning('Optimization problem status failure')
+                logging.warning('Optimization problem status failure')
             beta_sol = np.concatenate([b.value for b in beta_var], axis=0)
             beta_sol[np.abs(beta_sol) < self.tol] = 0
             beta_sol_list.append(beta_sol)
@@ -388,8 +417,9 @@ class ASGL:
         """
         n = x.shape[0]
         # Check th group_index, find the unique groups, count how many vars are in each group (this is the group size)
+        group_index = np.asarray(group_index).astype(int)
         unique_group_index = np.unique(group_index)
-        group_sizes, beta_var = self.__num_beta_var_from_group_index(group_index)
+        group_sizes, beta_var = self._num_beta_var_from_group_index(group_index)
         num_groups = len(group_sizes)
         model_prediction = 0
         lasso_penalization = 0
@@ -413,7 +443,7 @@ class ASGL:
         if self.model == 'lm':
             objective_function = (1.0 / n) * cvxpy.sum_squares(y - model_prediction)
         else:
-            objective_function = (1.0 / n) * cvxpy.sum(self.__quantile_function(x=(y - model_prediction)))
+            objective_function = (1.0 / n) * cvxpy.sum(self._quantile_function(x=(y - model_prediction)))
         lasso_param = cvxpy.Parameter(nonneg=True)
         grp_lasso_param = cvxpy.Parameter(nonneg=True)
         objective = cvxpy.Minimize(objective_function +
@@ -425,21 +455,31 @@ class ASGL:
         for lam, al in param:
             lasso_param.value = lam * al
             grp_lasso_param.value = lam * (1 - al)
-            # Solve the problem. Try first default CVXPY option, which is usually optimal for the problem. If a
-            # ValueError arises, try the solvers provided as input to the method.
+            # Solve the problem. If solver is left as default, try optimal solver sugested by cvxpy.
+            # If other name is provided, try the name provided
+            # If these options fail, try default ECOS, OSQP, SCS options
             try:
-                problem.solve(warm_start=True)
+                if self.solver == 'default':
+                    problem.solve(warm_start=True)
+                else:
+                    solver_dict = self._cvxpy_solver_options(solver=self.solver)
+                    problem.solve(**solver_dict)
             except (ValueError, cvxpy.error.SolverError):
-                for elt in self.solver:
-                    solver_dict = self.__cvxpy_solver_options(solver=elt)
+                logging.warning(
+                    'Default solver failed. Using alternative options. Check solver and solver_stats for more '
+                    'details')
+                solver = ['ECOS', 'OSQP', 'SCS']
+                for elt in solver:
+                    solver_dict = self._cvxpy_solver_options(solver=elt)
                     try:
                         problem.solve(**solver_dict)
                         if 'optimal' in problem.status:
                             break
                     except (ValueError, cvxpy.error.SolverError):
                         continue
+            self.solver_stats = problem.solver_stats
             if problem.status in ["infeasible", "unbounded"]:
-                logger.warning('Optimization problem status failure')
+                logging.warning('Optimization problem status failure')
             beta_sol = np.concatenate([b.value for b in beta_var], axis=0)
             beta_sol[np.abs(beta_sol) < self.tol] = 0
             beta_sol_list.append(beta_sol)
@@ -465,32 +505,42 @@ class ASGL:
         if self.model == 'lm':
             objective_function = (1.0 / n) * cvxpy.sum_squares(y - x @ beta_var)
         else:
-            objective_function = (1.0 / n) * cvxpy.sum(self.__quantile_function(x=(y - x @ beta_var)))
+            objective_function = (1.0 / n) * cvxpy.sum(self._quantile_function(x=(y - x @ beta_var)))
         objective = cvxpy.Minimize(objective_function + lasso_penalization)
         problem = cvxpy.Problem(objective)
         beta_sol_list = []
         # Solve the problem iteratively for each parameter value
         for lam, lw in param:
             l_weights_param.value = lam * lw
-            # Solve the problem. Try first default CVXPY option, which is usually optimal for the problem. If a
-            # ValueError arises, try the solvers provided as input to the method.
+            # Solve the problem. If solver is left as default, try optimal solver sugested by cvxpy.
+            # If other name is provided, try the name provided
+            # If these options fail, try default ECOS, OSQP, SCS options
             try:
-                problem.solve(warm_start=True)
+                if self.solver == 'default':
+                    problem.solve(warm_start=True)
+                else:
+                    solver_dict = self._cvxpy_solver_options(solver=self.solver)
+                    problem.solve(**solver_dict)
             except (ValueError, cvxpy.error.SolverError):
-                for elt in self.solver:
-                    solver_dict = self.__cvxpy_solver_options(solver=elt)
+                logging.warning(
+                    'Default solver failed. Using alternative options. Check solver and solver_stats for more '
+                    'details')
+                solver = ['ECOS', 'OSQP', 'SCS']
+                for elt in solver:
+                    solver_dict = self._cvxpy_solver_options(solver=elt)
                     try:
                         problem.solve(**solver_dict)
                         if 'optimal' in problem.status:
                             break
                     except (ValueError, cvxpy.error.SolverError):
                         continue
+            self.solver_stats = problem.solver_stats
             if problem.status in ["infeasible", "unbounded"]:
-                logger.warning('Optimization problem status failure')
+                logging.warning('Optimization problem status failure')
             beta_sol = beta_var.value
             beta_sol[np.abs(beta_sol) < self.tol] = 0
             beta_sol_list.append(beta_sol)
-        logger.debug('Function finished without errors')
+        logging.debug('Function finished without errors')
         return beta_sol_list
 
     def agl(self, x, y, group_index, param):
@@ -499,8 +549,9 @@ class ASGL:
         """
         n = x.shape[0]
         # Check th group_index, find the unique groups, count how many vars are in each group (this is the group size)
+        group_index = np.asarray(group_index).astype(int)
         unique_group_index = np.unique(group_index)
-        group_sizes, beta_var = self.__num_beta_var_from_group_index(group_index)
+        group_sizes, beta_var = self._num_beta_var_from_group_index(group_index)
         num_groups = len(group_sizes)
         model_prediction = 0
         group_lasso_penalization = 0
@@ -525,28 +576,38 @@ class ASGL:
         if self.model == 'lm':
             objective_function = (1.0 / n) * cvxpy.sum_squares(y - model_prediction)
         else:
-            objective_function = (1.0 / n) * cvxpy.sum(self.__quantile_function(x=(y - model_prediction)))
+            objective_function = (1.0 / n) * cvxpy.sum(self._quantile_function(x=(y - model_prediction)))
         objective = cvxpy.Minimize(objective_function + group_lasso_penalization)
         problem = cvxpy.Problem(objective)
         beta_sol_list = []
         # Solve the problem iteratively for each parameter value
         for lam, gl in param:
             gl_weights_param.value = lam * gl
-            # Solve the problem. Try first default CVXPY option, which is usually optimal for the problem. If a
-            # ValueError arises, try the solvers provided as input to the method.
+            # Solve the problem. If solver is left as default, try optimal solver sugested by cvxpy.
+            # If other name is provided, try the name provided
+            # If these options fail, try default ECOS, OSQP, SCS options
             try:
-                problem.solve(warm_start=True)
+                if self.solver == 'default':
+                    problem.solve(warm_start=True)
+                else:
+                    solver_dict = self._cvxpy_solver_options(solver=self.solver)
+                    problem.solve(**solver_dict)
             except (ValueError, cvxpy.error.SolverError):
-                for elt in self.solver:
-                    solver_dict = self.__cvxpy_solver_options(solver=elt)
+                logging.warning(
+                    'Default solver failed. Using alternative options. Check solver and solver_stats for more '
+                    'details')
+                solver = ['ECOS', 'OSQP', 'SCS']
+                for elt in solver:
+                    solver_dict = self._cvxpy_solver_options(solver=elt)
                     try:
                         problem.solve(**solver_dict)
                         if 'optimal' in problem.status:
                             break
                     except (ValueError, cvxpy.error.SolverError):
                         continue
+            self.solver_stats = problem.solver_stats
             if problem.status in ["infeasible", "unbounded"]:
-                logger.warning('Optimization problem status failure')
+                logging.warning('Optimization problem status failure')
             beta_sol = np.concatenate([b.value for b in beta_var], axis=0)
             beta_sol[np.abs(beta_sol) < self.tol] = 0
             beta_sol_list.append(beta_sol)
@@ -558,8 +619,9 @@ class ASGL:
         """
         n, m = x.shape
         # Check th group_index, find the unique groups, count how many vars are in each group (this is the group size)
+        group_index = np.asarray(group_index).astype(int)
         unique_group_index = np.unique(group_index)
-        group_sizes, beta_var = self.__num_beta_var_from_group_index(group_index)
+        group_sizes, beta_var = self._num_beta_var_from_group_index(group_index)
         num_groups = len(group_sizes)
         model_prediction = 0
         alasso_penalization = 0
@@ -587,7 +649,7 @@ class ASGL:
         if self.model == 'lm':
             objective_function = (1.0 / n) * cvxpy.sum_squares(y - model_prediction)
         else:
-            objective_function = (1.0 / n) * cvxpy.sum(self.__quantile_function(x=(y - model_prediction)))
+            objective_function = (1.0 / n) * cvxpy.sum(self._quantile_function(x=(y - model_prediction)))
         objective = cvxpy.Minimize(objective_function +
                                    a_group_lasso_penalization +
                                    alasso_penalization)
@@ -597,21 +659,31 @@ class ASGL:
         for lam, al, lw, glw in param:
             l_weights_param.value = lw * lam * al
             gl_weights_param.value = glw * lam * (1 - al)
-            # Solve the problem. Try first default CVXPY option, which is usually optimal for the problem. If a
-            # ValueError arises, try the solvers provided as input to the method.
+            # Solve the problem. If solver is left as default, try optimal solver sugested by cvxpy.
+            # If other name is provided, try the name provided
+            # If these options fail, try default ECOS, OSQP, SCS options
             try:
-                problem.solve(warm_start=True)
+                if self.solver == 'default':
+                    problem.solve(warm_start=True)
+                else:
+                    solver_dict = self._cvxpy_solver_options(solver=self.solver)
+                    problem.solve(**solver_dict)
             except (ValueError, cvxpy.error.SolverError):
-                for elt in self.solver:
-                    solver_dict = self.__cvxpy_solver_options(solver=elt)
+                logging.warning(
+                    'Default solver failed. Using alternative options. Check solver and solver_stats for more '
+                    'details')
+                solver = ['ECOS', 'OSQP', 'SCS']
+                for elt in solver:
+                    solver_dict = self._cvxpy_solver_options(solver=elt)
                     try:
                         problem.solve(**solver_dict)
                         if 'optimal' in problem.status:
                             break
                     except (ValueError, cvxpy.error.SolverError):
                         continue
+            self.solver_stats = problem.solver_stats
             if problem.status in ["infeasible", "unbounded"]:
-                logger.warning('Optimization problem status failure')
+                logging.warning('Optimization problem status failure')
             beta_sol = np.concatenate([b.value for b in beta_var], axis=0)
             beta_sol[np.abs(beta_sol) < self.tol] = 0
             beta_sol_list.append(beta_sol)
@@ -619,7 +691,7 @@ class ASGL:
 
     # PARALLEL CODE ###################################################################################################
 
-    def parallel_execution(self, x, y, param, group_index=None):
+    def _parallel_execution(self, x, y, param, group_index=None):
         """
         Parallel implementation of the solvers
         """
@@ -643,9 +715,9 @@ class ASGL:
         # Solve problem in parallel
         pool = mp.Pool(num_chunks)
         if self.penalization in ['lasso', 'alasso']:
-            global_results = pool.map(functools.partial(getattr(self, self.__get_solver_names()), x, y), chunks)
+            global_results = pool.map(functools.partial(getattr(self, self._get_solver_names()), x, y), chunks)
         else:
-            global_results = pool.map(functools.partial(getattr(self, self.__get_solver_names()), x, y, group_index),
+            global_results = pool.map(functools.partial(getattr(self, self._get_solver_names()), x, y, group_index),
                                       chunks)
         pool.close()
         pool.join()
@@ -661,7 +733,7 @@ class ASGL:
 
     # FIT METHOD ######################################################################################################
 
-    def __get_solver_names(self):
+    def _get_solver_names(self):
         if 'asgl' in self.penalization:
             return 'asgl'
         else:
@@ -672,18 +744,18 @@ class ASGL:
         Main function of the module. Given a model, penalization and parameter values specified in the class definition,
         this function solves the model and produces the regression coefficients
         """
-        param = self.__preprocessing()
+        param = self._preprocessing()
         if self.penalization is None:
             self.coef_ = self.unpenalized_solver(x=x, y=y)
         else:
             if self.parallel is False:
                 if self.penalization in ['lasso', 'alasso']:
-                    self.coef_ = getattr(self, self.__get_solver_names())(x=x, y=y, param=param)
+                    self.coef_ = getattr(self, self._get_solver_names())(x=x, y=y, param=param)
                 else:
-                    self.coef_ = getattr(self, self.__get_solver_names())(x=x, y=y, param=param,
-                                                                          group_index=group_index)
+                    self.coef_ = getattr(self, self._get_solver_names())(x=x, y=y, param=param,
+                                                                         group_index=group_index)
             else:
-                self.coef_ = self.parallel_execution(x=x, y=y, param=param, group_index=group_index)
+                self.coef_ = self._parallel_execution(x=x, y=y, param=param, group_index=group_index)
 
     # PREDICTION METHOD ###############################################################################################
 
@@ -695,7 +767,7 @@ class ASGL:
         if self.intercept:
             x_new = np.c_[np.ones(x_new.shape[0]), x_new]
         if x_new.shape[1] != len(self.coef_[0]):
-            logger.error('Model dimension and new data dimension does not match')
+            logging.error('Model dimension and new data dimension does not match')
             raise ValueError('Model dimension and new data dimension does not match')
         # Store predictions in a list
         prediction_list = []
@@ -716,17 +788,21 @@ class ASGL:
         - n_gl_weights: number of different weights for the lasso part of the asgl (or asgl_gl) penalizations
         """
         # Run the input_checker to verify that the inputs have the correct format
-        if self.__input_checker() is False:
-            logger.error('incorrect input parameters')
+        if self._input_checker() is False:
+            logging.error('incorrect input parameters')
             raise ValueError('incorrect input parameters')
         if self.penalization is None:
             # See meaning of each element in the "else" result statement.
-            result = [1, None, None, None, None]
+            result = dict(num_models=1,
+                          n_lambda=None,
+                          n_alpha=None,
+                          n_lasso_weights=None,
+                          n_gl_weights=None)
         else:
-            n_lambda, drop = self.__preprocessing_lambda()
-            n_alpha, drop = self.__preprocessing_alpha()
-            n_lasso_weights, drop = self.__preprocessing_weights(self.lasso_weights)
-            n_gl_weights, drop = self.__preprocessing_weights(self.gl_weights)
+            n_lambda, drop = self._preprocessing_lambda()
+            n_alpha, drop = self._preprocessing_alpha()
+            n_lasso_weights, drop = self._preprocessing_weights(self.lasso_weights)
+            n_gl_weights, drop = self._preprocessing_weights(self.gl_weights)
             list_param = [n_lambda, n_alpha, n_lasso_weights, n_gl_weights]
             list_param_no_none = [elt for elt in list_param if elt is not None]
             num_models = np.prod(list_param_no_none)
@@ -739,7 +815,7 @@ class ASGL:
 
     def _retrieve_parameters_idx(self, param_index):
         """
-        Given an index for the param iterable output from __preprocessing function, this function returns a tupple
+        Given an index for the param iterable output from _preprocessing function, this function returns a tupple
         with the index of the value of each parameter.
         Example: Solving an adaptive sparse group lasso model with 5 values for lambda1, 4 values for alpha,
                  3 possible lasso weights and 3 possible group lasso weights yields in a grid search on
@@ -755,7 +831,7 @@ class ASGL:
         if param_index > n_models:
             string = f'param_index should be smaller or equal than the number of models solved. n_models={n_models}, ' \
                      f'param_index={param_index}'
-            logger.error(string)
+            logging.error(string)
             raise ValueError(string)
         # If penalization is None, all parameters are set to None
         if self.penalization is None:
@@ -801,7 +877,7 @@ class ASGL:
 
 # ERROR CALCULATOR METHOD #############################################################################################
 
-def __quantile_function(y_true, y_pred, tau):
+def _quantile_function(y_true, y_pred, tau):
     """
     Quantile function required for error computation
     """
@@ -816,13 +892,13 @@ def error_calculator(y_true, prediction_list, error_type="MSE", tau=None):
         MSE=mean_squared_error,
         MAE=mean_absolute_error,
         MDAE=median_absolute_error,
-        QRE=__quantile_function)
+        QRE=_quantile_function)
     valid_error_types = error_dict.keys()
     # Check that the error_type is a valid error type considered
     if error_type not in valid_error_types:
         raise ValueError(f'invalid error type. Valid error types are {error_dict.keys()}')
     if y_true.shape[0] != len(prediction_list[0]):
-        logger.error('Dimension of test data does not match dimension of prediction')
+        logging.error('Dimension of test data does not match dimension of prediction')
         raise ValueError('Dimension of test data does not match dimension of prediction')
     # For each prediction, store the error associated to that prediction in a list
     error_list = []
