@@ -1,12 +1,13 @@
 import logging
+import sys
 
 import numpy as np
-from sklearn.model_selection import KFold
+from sklearn.model_selection import KFold, GroupKFold
 
 from . import asgl
 from . import weights
 
-logger = logging.getLogger(__name__)
+logging.basicConfig(stream=sys.stdout, level=logging.INFO, format='%(asctime)s - %(levelname)s - %(message)s')
 
 
 class CvGeneralClass(asgl.ASGL):
@@ -71,27 +72,38 @@ class CV(CvGeneralClass):
             random_state: random state value in case reproducible data splits are required
             nfolds: number of folds in which the dataset should be split. Default value is 5
         """
-        # ASGL
+        # CvGeneralClass
         super().__init__(model, penalization, intercept, tol, lambda1, alpha, tau, lasso_weights, gl_weights, parallel,
-                         num_cores, solver, max_iters, weight_technique, weight_tol, lasso_power_weight, gl_power_weight,
-                         variability_pct, lambda1_weights, spca_alpha, spca_ridge_alpha, error_type, random_state)
+                         num_cores, solver, max_iters, weight_technique, weight_tol, lasso_power_weight,
+                         gl_power_weight, variability_pct, lambda1_weights, spca_alpha, spca_ridge_alpha, error_type,
+                         random_state)
         # Relative to cross validation / train validate / test
         self.nfolds = nfolds
 
     # SPLIT DATA METHODS ##############################################################################################
 
-    def __cross_validation_split(self, nrows):
-        # Randomly generate index
-        data_index = np.random.choice(nrows, nrows, replace=False)
-        # Split data into k folds
-        k_folds = KFold(n_splits=self.nfolds).split(data_index)
+    def _cross_validation_split(self, nrows, split_index=None):
+        """
+        Split data based on kfold or group kfold cross validation
+        :param nrows: number of rows in the dataset
+        :param split_index: Group structure of observations used in GroupKfold. 
+                            same length as nrows
+        """
+        if split_index is None:
+            # Randomly generate index
+            data_index = np.random.choice(nrows, nrows, replace=False)
+            # Split data into k folds
+            k_folds = KFold(n_splits=self.nfolds).split(data_index)
+        else:
+            data_index = np.arange(0, nrows)
+            k_folds = GroupKFold(n_splits=self.nfolds).split(X=data_index, groups=split_index)
         # List containing zips of (train, test) indices
         response = [(data_index[train], data_index[validate]) for train, validate in list(k_folds)]
         return response
 
     # CROSS VALIDATION ################################################################################################
 
-    def __one_step_cross_validation(self, x, y, group_index=None, zip_index=None):
+    def _one_step_cross_validation(self, x, y, group_index=None, zip_index=None):
         # Given a zip (one element from the cross_validation_split function) retrieve train / test splits
         train_index, test_index = zip_index
         x_train, x_test = (x[train_index, :], x[test_index, :])
@@ -102,14 +114,14 @@ class CV(CvGeneralClass):
                                       tau=self.tau)
         return error
 
-    def cross_validation(self, x, y, group_index=None):
+    def cross_validation(self, x, y, group_index=None, split_index=None):
         error_list = []
         # Define random state if required
         if self.random_state is not None:
             np.random.seed(self.random_state)
-        cv_index = self.__cross_validation_split(nrows=x.shape[0])
+        cv_index = self._cross_validation_split(nrows=x.shape[0], split_index=split_index)
         for zip_index in cv_index:
-            error = self.__one_step_cross_validation(x, y, group_index=group_index, zip_index=zip_index)
+            error = self._one_step_cross_validation(x, y, group_index=group_index, zip_index=zip_index)
             error_list.append(error)
         return np.array(error_list).T  # Each row is a model, each column is a k fold split
 
@@ -122,8 +134,9 @@ class TVT(CvGeneralClass):
                  random_state=None, train_pct=0.05, validate_pct=0.05, train_size=None, validate_size=None):
 
         super().__init__(model, penalization, intercept, tol, lambda1, alpha, tau, lasso_weights, gl_weights, parallel,
-                         num_cores, solver, max_iters, weight_technique, weight_tol, lasso_power_weight, gl_power_weight,
-                         variability_pct, lambda1_weights, spca_alpha, spca_ridge_alpha, error_type, random_state)
+                         num_cores, solver, max_iters, weight_technique, weight_tol, lasso_power_weight,
+                         gl_power_weight, variability_pct, lambda1_weights, spca_alpha, spca_ridge_alpha, error_type,
+                         random_state)
         # Relative to / train validate / test
         self.train_pct = train_pct
         self.validate_pct = validate_pct
@@ -132,7 +145,12 @@ class TVT(CvGeneralClass):
 
     # TRAIN VALIDATE TEST SPLIT #######################################################################################
 
-    def __train_validate_test_split(self, nrows):
+    def _train_validate_test_split(self, nrows):
+        """
+        Splits data into train validate and test. Takes into account the random state for consistent repetitions
+        :param nrows: Number of observations
+        :return:  A three items object including train index, validate index and test index
+        """
         # Randomly generate index
         data_index = np.random.choice(nrows, nrows, replace=False)
         if self.train_size is None:
@@ -145,12 +163,20 @@ class TVT(CvGeneralClass):
 
     # TRAIN VALIDATE TEST #############################################################################################
 
-    def __train_validate(self, x, y, group_index=None):
+    def _train_validate(self, x, y, group_index=None):
+        """
+        Split the data into train validate and test. Fit an adaptive lasso based model on the train split. Obtain
+        predictions and compute error using validation set.
+        :param x: data matrix x
+        :param y: response vector y
+        :param group_index: group index of predictors in data matrix x
+        :return: Validate error and test index
+        """
         # Define random state if required
         if self.random_state is not None:
             np.random.seed(self.random_state)
         # Split data
-        split_index = self.__train_validate_test_split(nrows=x.shape[0])
+        split_index = self._train_validate_test_split(nrows=x.shape[0])
         x_train, x_validate, drop = [x[idx, :] for idx in split_index]
         y_train, y_validate, drop = [y[idx] for idx in split_index]
         test_index = split_index[2]
@@ -162,7 +188,16 @@ class TVT(CvGeneralClass):
                                                error_type=self.error_type, tau=self.tau)
         return validate_error, test_index
 
-    def __tv_test(self, x, y, validate_error, test_index):
+    def _tv_test(self, x, y, validate_error, test_index):
+        """
+        Given a validate error and a test index obtains the final test_error and stores the optimal parameter values and
+        optimal betas
+        :param x: data matrix x
+        :param y: response vector y
+        :param validate_error: Validation error computed in __train_validate()
+        :param test_index: Test index
+        :return: optimal_betas, optimal_parameters, test_error
+        """
         # Select the minimum error
         minimum_error_idx = np.argmin(validate_error)
         # Select the parameters index associated to mininum error values
@@ -171,12 +206,18 @@ class TVT(CvGeneralClass):
         optimal_betas = self.coef_[minimum_error_idx]
         test_prediction = [self.predict(x_new=x[test_index, :])[minimum_error_idx]]
         test_error = asgl.error_calculator(y_true=y[test_index], prediction_list=test_prediction,
-                                           error_type=self.error_type, tau=self.tau)
+                                           error_type=self.error_type, tau=self.tau)[0]
         return optimal_betas, optimal_parameters, test_error
 
     def train_validate_test(self, x, y, group_index=None):
-        validate_error, test_index = self.__train_validate(x, y, group_index)
-        optimal_betas, optimal_parameters, test_error = self.__tv_test(x, y, validate_error, test_index)
+        """
+        Runs functions __train_validate() and __tv_test(). Stores results in a dictionary
+        :param x: data matrix x
+        :param y: response vector y
+        :param group_index: group index of predictors in data matrix x
+        """
+        validate_error, test_index = self._train_validate(x, y, group_index)
+        optimal_betas, optimal_parameters, test_error = self._tv_test(x, y, validate_error, test_index)
         result = dict(
             optimal_betas=optimal_betas,
             optimal_parameters=optimal_parameters,
@@ -187,6 +228,9 @@ class TVT(CvGeneralClass):
 # TRAIN TEST SPLIT ###################################################################################################
 
 def train_test_split(nrows, train_size=None, train_pct=0.7, random_state=None):
+    """
+    Splits data into train / test. Takes into account random_state for future consistency.
+    """
     # Define random state if required
     if random_state is not None:
         np.random.seed(random_state)
@@ -195,7 +239,7 @@ def train_test_split(nrows, train_size=None, train_pct=0.7, random_state=None):
         train_size = int(round(nrows * train_pct))
     # Check that nrows is larger than train_size
     if nrows < train_size:
-        logger.error(f'Train size is too large. Input number of rows:{nrows}, current train_size: {train_size}')
+        logging.error(f'Train size is too large. Input number of rows:{nrows}, current train_size: {train_size}')
     # List of 2 elements of size train_size, remaining_size (test)
     split_index = np.split(data_index, [train_size])
     train_idx, test_idx = [elt for elt in split_index]
